@@ -1,9 +1,9 @@
 #include <sys/queue.h>
-#ifdef _ALLOCA_H
+/* Alloca is defined in stdlib.h in NetBSD */
+#ifndef __NetBSD__
 #include <alloca.h>
-#else
-#include <limits.h>
 #endif
+#include <limits.h>
 #include <ctype.h>
 #include <dirent.h>
 #include <errno.h>
@@ -41,6 +41,10 @@
 #endif
 #endif
 
+#ifdef __hpux
+#include "asprintf.c"
+#endif
+
 #if !defined(asprintf)
 #include <stdarg.h>
 
@@ -74,6 +78,29 @@ int asprintf(char **string_ptr, const char *format, ...)
 
 #endif
 
+#if !defined(strndup)
+char *strndup(const char *s, size_t n)
+{
+       size_t nAvail;
+       char *p;
+
+       if(!s)
+               return NULL;
+
+       /* min() */
+       nAvail = strlen(s) + 1;
+       if ( (n + 1) < nAvail)
+               nAvail = n + 1;
+
+       p = malloc(nAvail);
+       if (!p)
+               return NULL;
+       memcpy(p, s, nAvail);
+       p[nAvail - 1] = 0;
+       return p;
+}
+#endif
+
 enum {
 	STATE_DEFAULT = 2,
 	STATE_SKIP_LINE = 4,
@@ -166,7 +193,7 @@ static char *readPath(const char *configFile, int lineNum, char *key,
 
 	chptr = start;
 
-	while( (len = mbrtowc(&pwc, chptr, strlen(chptr), NULL)) != 0 ) {
+	while( (len = mbrtowc(&pwc, chptr, strlen(chptr), NULL)) != 0 && strlen(chptr) != 0) {
 		if( len == (size_t)(-1) || len == (size_t)(-2) || !iswprint(pwc) || iswblank(pwc) ) {
 		    message(MESS_ERROR, "%s:%d bad %s path %s\n",
 			    configFile, lineNum, key, start);
@@ -270,6 +297,7 @@ static void copyLogInfo(struct logInfo *to, struct logInfo *from)
     to->criterium = from->criterium;
     to->threshhold = from->threshhold;
     to->minsize = from->minsize;
+	to->maxsize = from->maxsize;
     to->rotateCount = from->rotateCount;
     to->rotateAge = from->rotateAge;
     to->logStart = from->logStart;
@@ -292,6 +320,7 @@ static void copyLogInfo(struct logInfo *to, struct logInfo *from)
     if (from->compress_ext)
 	to->compress_ext = strdup(from->compress_ext);
     to->flags = from->flags;
+	to->shred_cycles = from->shred_cycles;
     to->createMode = from->createMode;
     to->createUid = from->createUid;
     to->createGid = from->createGid;
@@ -484,6 +513,7 @@ int readAllConfigPaths(const char **paths)
 		.criterium = ROT_SIZE,
 		.threshhold = 1024 * 1024,
 		.minsize = 0,
+		.maxsize = 0,
 		.rotateCount = 0,
 		.rotateAge = 0,
 		.logStart = -1,
@@ -551,7 +581,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
     char foo;
     off_t length;
     int lineNum = 1;
-    int multiplier;
+    unsigned long long multiplier;
     int i, k;
     char *scriptStart = NULL;
     char **scriptDest = NULL;
@@ -573,6 +603,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
     struct logInfo *log;
 	static unsigned recursion_depth = 0U;
 	char *globerr_msg = NULL;
+	int in_config = 0;
 	struct flock fd_lock = {
 		.l_start = 0,
 		.l_len = 0,
@@ -817,6 +848,17 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 
 					rc = sscanf(key, "%o %200s %200s%c", &createMode,
 							createOwner, createGroup, &foo);
+					/* We support 'create <owner> <group> notation now */
+					if (rc == 0) {
+						rc = sscanf(key, "%200s %200s%c",
+								createOwner, createGroup, &foo);
+						/* Simulate that we have read createMode and se it
+						 * to NO_MODE. */
+						if (rc > 0) {
+							createMode = NO_MODE;
+							rc += 1;
+						}
+					}
 					if (rc == 4) {
 						message(MESS_ERROR, "%s:%d extra arguments for "
 							"create\n", configFile, lineNum);
@@ -865,13 +907,13 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					newlog->flags |= LOG_FLAG_CREATE;
 				} else if (!strcmp(key, "nocreate")) {
 					newlog->flags &= ~LOG_FLAG_CREATE;
-				} else if (!strcmp(key, "size") || !strcmp(key, "minsize")) {
+				} else if (!strcmp(key, "size") || !strcmp(key, "minsize") ||
+							!strcmp(key, "maxsize")) {
 					unsigned long long size = 0;
 					char *opt = key;
 							
 					if ((key = isolateValue(configFile, lineNum, opt, &start,
 							&buf, length)) != NULL) {
-						free(opt);
 						int l = strlen(key) - 1;
 						if (key[l] == 'k') {
 							key[l] = '\0';
@@ -883,6 +925,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							key[l] = '\0';
 							multiplier = 1024 * 1024 * 1024;
 						} else if (!isdigit(key[l])) {
+							free(opt);
 							message(MESS_ERROR, "%s:%d unknown unit '%c'\n",
 								configFile, lineNum, key[l]);
 							if (newlog != defConfig) {
@@ -899,6 +942,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						if (*chptr) {
 							message(MESS_ERROR, "%s:%d bad size '%s'\n",
 								configFile, lineNum, key);
+							free(opt);
 							if (newlog != defConfig) {
 								state = STATE_ERROR;
 								continue;
@@ -906,11 +950,15 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 								goto error;
 							}
 						}
-						if (!strncmp(key, "size", 4)) {
+						if (!strncmp(opt, "size", 4)) {
 						  newlog->criterium = ROT_SIZE;
 						  newlog->threshhold = size;
-						} else
+						} else if (!strncmp(opt, "maxsize", 7)) {
+						  newlog->maxsize = size;
+						} else {
 						  newlog->minsize = size;
+						}
+						free(opt);
 					}
 					else {
 						free(opt);
@@ -1112,7 +1160,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							goto error;
 						}
 					}
-					else continue;
+
 #if 0
 					if (stat(newlog->oldDir, &sb)) {
 						message(MESS_ERROR, "%s:%d error verifying olddir "
@@ -1157,7 +1205,6 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							goto error;
 						}
 					}
-					else continue;
 
 					if (access(newlog->compress_prog, X_OK)) {
 						message(MESS_ERROR,
@@ -1188,7 +1235,6 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 							goto error;
 						}
 					}
-					else continue;
 
 					if (access(newlog->uncompress_prog, X_OK)) {
 						message(MESS_ERROR,
@@ -1214,17 +1260,14 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						newlog->compress_options_count = 0;
 					}
 
-					if (!
-						(options =
-							readPath(configFile, lineNum, "compressoptions",
-								&start, &buf, length))) {
+					if (!(options = isolateLine(&start, &buf, length))) {
 						if (newlog != defConfig) {
 							state = STATE_ERROR;
 							continue;
 						} else {
 							goto error;
 						}
-					} else continue;
+					}
 
 					if (poptParseArgvString(options,
 								&newlog->compress_options_count,
@@ -1257,7 +1300,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						} else {
 							goto error;
 						}
-					} else continue;
+					}
 
 					message(MESS_DEBUG, "compress_ext is now %s\n",
 						newlog->compress_ext);
@@ -1269,7 +1312,12 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 				}
 				free(key);
 				key = NULL;
-			} else if (*start == '/' || *start == '"' || *start == '\'') {
+			} else if (*start == '/' || *start == '"' || *start == '\''
+#ifdef GLOB_TILDE
+                                                                           || *start == '~'
+#endif
+                                                                           ) {
+				in_config = 0;
 				if (newlog != defConfig) {
 					message(MESS_ERROR, "%s:%d unexpected log filename\n",
 						configFile, lineNum);
@@ -1292,10 +1340,23 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					goto error;
 
 				endtag = start;
-				while (endtag - buf < length && *endtag != '{' && *endtag != '\0') {
+				while (endtag - buf < length && *endtag != '{' && *endtag != '}' && *endtag != '\0') {
 					endtag++;}
 				if (endtag - buf > length)
 					continue;
+				if (*endtag == '}') {
+					message(MESS_ERROR, "%s:%d unexpected } (missing previous '{')\n", configFile,
+						lineNum);
+					goto error;
+				}
+				if (*endtag == '{') {
+					in_config = 1;
+				}
+				else {
+					message(MESS_ERROR, "%s:%d missing '{' after log files definition\n", configFile,
+						lineNum);
+					goto error;
+				}
 				char *key = strndup(start, endtag - start);
 				start = endtag;
 
@@ -1320,8 +1381,11 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 					globerr_msg = NULL;
 				}
 					
-				rc = glob(argv[argNum], GLOB_NOCHECK, globerr,
-					&globResult);
+				rc = glob(argv[argNum], GLOB_NOCHECK
+#ifdef GLOB_TILDE
+                                                        | GLOB_TILDE
+#endif 
+                                                    , globerr, &globResult);
 				if (rc == GLOB_ABORTED) {
 					if (newlog->flags & LOG_FLAG_MISSINGOK) {
 						continue;
@@ -1387,6 +1451,12 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 						lineNum);
 					goto error;
 				}
+				if (!in_config) {
+					message(MESS_ERROR, "%s:%d unexpected } (missing previous '{')\n", configFile,
+						lineNum);
+					goto error;
+				}
+				in_config = 0;
 			if (globerr_msg) {
 				if (!(newlog->flags & LOG_FLAG_MISSINGOK))
 					message(MESS_ERROR, "%s", globerr_msg);
@@ -1540,7 +1610,7 @@ static int readConfigFile(const char *configFile, struct logInfo *defConfig)
 	}
 
     }
-
+ 
     if (scriptStart) {
 	message(MESS_ERROR,
 		"%s:prerotate or postrotate without endscript\n",
